from abc import ABC, abstractmethod
from concurrent.futures import (Executor, ProcessPoolExecutor,
                                ThreadPoolExecutor)
from datetime import datetime
from typing import Any, Union

import pandas as pd

from nure.debug import DEBUG
from nure.sync.s3 import S3, S3Uri


class VolatileObject(ABC):
    _default_pool_ = None

    def __init__(self, pool: Executor = None) -> None:
        super().__init__()
        if pool is None:
            if VolatileObject._default_pool_ is None:
                VolatileObject._default_pool_ = ThreadPoolExecutor() if DEBUG else ProcessPoolExecutor()
            pool = VolatileObject._default_pool_

        self.pool = pool
        self.future = None
        self.instance = None

    @abstractmethod
    def is_valid(self, *args: Any, **kwds: Any):
        raise NotImplementedError()

    @abstractmethod
    def retrieve(self, *args: Any, **kwds: Any) -> Any:
        raise NotImplementedError()

    def get_instance(self, *args: Any, force=False, **kwds: Any) -> Any:
        if self.future is None and (force or not self.is_valid(*args, **kwds)):
            self.future = self.pool.submit(self.retrieve, *args, **kwds)

        if self.future is not None and (force or self.future.done()):
            self.instance = self.future.result()
            self.future = None

        return self.instance


class TimeoutObject(VolatileObject):
    def __init__(self, ttl: float, pool: Executor = None) -> None:
        self.ttl = ttl
        self.mod_time = datetime.min
        super().__init__(pool)

    def is_valid(self):
        return (datetime.now() - self.mod_time).total_seconds() < self.ttl

    @abstractmethod
    def retrieve(self) -> Any:
        self.mod_time = datetime.now()
        return None


class S3Object(VolatileObject):
    def __init__(self, s3uri: Union[str, S3Uri],
                 s3: S3 = S3(), ttc: float = 60,
                 pool: Executor = None) -> None:
        self.s3uri = s3uri
        self.s3 = s3
        self.ttc = ttc
        self.last_check = datetime.min
        self.last_mod = datetime.min
        super().__init__(pool=pool)

    def get_last_mod(self) -> datetime:
        return self.s3.metadata(self.s3uri)['LastModified']

    def is_valid(self):
        # if not time to check
        if (datetime.now() - self.last_check).total_seconds() < self.ttc:
            return True

        last_mod = self.get_last_mod()
        self.last_check = datetime.now()
        return last_mod == self.last_mod

    def retrieve(self, ttl=1) -> Any:
        local_file = self.s3.require(self.s3uri, ttl=ttl)
        self.last_mod = self.get_last_mod()
        return local_file


class S3ParquetObject(S3Object):
    def retrieve(self, ttl=1) -> Any:
        local_file = super().retrieve(ttl=ttl)
        df = pd.read_parquet(local_file)
        return df



import csv
import os
import re
from abc import abstractmethod

import sqlalchemy
import nure.sync.cache


class Sql(nure.sync.cache.LocalFileCache):
    def __init__(self, suffix_func, root_path='data/sql', ttl=None) -> None:
        super(Sql, self).__init__(root_path, ttl)
        self.suffix_func = suffix_func

    @property
    @abstractmethod
    def engine(self) -> sqlalchemy.engine.Engine:
        raise NotImplementedError()

    def key_to_local_relative_path(self, key, *args, **kargs) -> str:
        fn, _ = os.path.splitext(os.path.basename(key))
        suffix = self.suffix_func(*args, **kargs)
        return f'{fn}{suffix}.csv'

    def retrieve(self, sql_key, local_file_path,
                 re_replace=None, sa_replace=None, partition_size=10000):
        with open(sql_key, 'rt') as fd:
            sql_str = fd.read()

        if isinstance(re_replace, dict):
            for pattern, repl in re_replace.items():
                sql_str = re.sub(pattern, repl, sql_str)

        with self.engine.connect().execution_options(stream_results=True) as conn:
            conn: sqlalchemy.engine.Connection
            result: sqlalchemy.engine.ResultProxy = conn.execute(sqlalchemy.text(sql_str), sa_replace or {})

            with open(local_file_path, 'wt', newline='') as csv_file:
                writer = csv.writer(csv_file, dialect='excel')
                writer.writerow(result.keys())

                while len(rows := result.fetchmany(size=partition_size)) > 0:
                    writer.writerows(rows)



import json

import sqlalchemy
import nure.sync.sql


class PostgreSql(nure.sync.sql.Sql):
    def __init__(self, secret_fn, suffix_func, root_path='data/postgre', ttl=None) -> None:
        super(PostgreSql, self).__init__(suffix_func, root_path, ttl)
        with open(secret_fn, 'rt') as fd:
            secret = json.load(fd)

        user = secret['user']
        password = secret['password']
        host = secret['host']
        port = secret['port']
        dbname = secret['dbname']

        self._engine = sqlalchemy.create_engine(
            f'postgresql://{user}:{password}@{host}:{port}/{dbname}')

    @property
    def engine(self) -> sqlalchemy.engine.Engine:
        return self._engine



from typing import Any, Dict, Iterable, Union
import more_itertools


class VentureCodeSuffix:
    def __init__(self, venture_code) -> None:
        self.venture_code = venture_code

    def __call__(self, *args, **kwds) -> str:
        return f'__{self.venture_code}'


class DailyActionLogSuffix:
    def __init__(self, venture_code) -> None:
        self.venture_code = venture_code

    def __call__(self, re_replace=None, *args, **kwds) -> str:
        if re_replace is None or (date := re_replace.get(R'\$\(date\)')) is None:
            return f'__{self.venture_code}'
        return f'__{date}__{self.venture_code}'


ReDict = Union[Iterable[str], Dict[str, 'ReDict']]


class ParameterSuffix:
    def __init__(self, suffix_config: ReDict) -> None:
        self.suffix_config = suffix_config

    @staticmethod
    def extract_param(suffix_config: ReDict, kwds: ReDict):
        if isinstance(suffix_config, dict):
            params = [
                ParameterSuffix.extract_param(suffix_config[key], kwds[key]) for key in suffix_config
            ]
            params = list(more_itertools.flatten(params))
        else:
            params = [kwds[key] for key in suffix_config]

        return params

    def __call__(self, *_: Any, **kwds: ReDict) -> Any:
        params = self.extract_param(self.suffix_config, kwds)
        if len(params) == 0:
            return ''

        params = [''] + params
        return '__'.join(params)



from abc import ABC, abstractmethod
import os
from datetime import datetime


class LocalFileCache(ABC):
    def __init__(self, root_path, ttl: float = None) -> None:
        self.root_path = root_path
        self.ttl = ttl

    def _is_valid_(self, local_file_path, ttl=None) -> bool:
        ttl = ttl or self.ttl
        if not os.path.isfile(local_file_path):
            return False

        if ttl is None:
            return True

        mod_time = datetime.fromtimestamp(os.stat(local_file_path).st_mtime)
        if (datetime.now() - mod_time).total_seconds() < ttl:
            return True

        return False

    def require(self, key, ttl=None, return_resynced=False, *args, **kargs):
        local_file_path = self.key_to_local_relative_path(key, *args, **kargs)
        local_file_path = os.path.join(self.root_path, local_file_path)
        is_resynced = False
        if not self._is_valid_(local_file_path, ttl=ttl):
            os.makedirs(os.path.dirname(local_file_path), exist_ok=True)
            self.retrieve(key, local_file_path, *args, **kargs)
            is_resynced = True

        if return_resynced:
            return local_file_path, is_resynced

        return local_file_path

    @abstractmethod
    def key_to_local_relative_path(self, key, *args, **kargs) -> str:
        raise NotImplementedError()

    @abstractmethod
    def retrieve(self, key, local_file_path, *args, **kargs) -> None:
        raise NotImplementedError()



from typing import Callable, Iterable


def chain(*l_func: Iterable[Callable]):
    def __wrapper__(argument):
        for func in l_func:
            argument = func(argument)

        return argument

    return __wrapper__


def from_iterable(func: Callable):
    def __wrapper__(iterable_arguments: Iterable):
        return func(*iterable_arguments)

    return __wrapper__



import collections.abc


def update(default, modified):
    for k, v in modified.items():
        if isinstance(v, collections.abc.Mapping):
            default[k] = update(default.get(k, {}), v)
        else:
            default[k] = v
    return default
